@@ -1,29 +1,48 @@
-#!/usr/local/bin/python2.7
-'''
+#!/usr/bin/env python3
+
+"""
 a2x - A toolchain manager for AsciiDoc (converts Asciidoc text files to other
       file formats)
 
-Copyright: Stuart Rackham (c) 2009
-License:   MIT
-Email:     srackham@gmail.com
+Free use of this software is granted under the terms of the MIT license.
 
-'''
+Copyright (C) 2002-2013 Stuart Rackham.
+Copyright (C) 2013-2020 AsciiDoc Contributors.
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in all
+copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+SOFTWARE.
+"""
 
 import os
 import fnmatch
-import HTMLParser
+from html.parser import HTMLParser
 import re
 import shutil
 import subprocess
 import sys
 import traceback
-import urlparse
+from urllib.parse import urlparse
 import zipfile
 import xml.dom.minidom
 import mimetypes
 
 PROG = os.path.basename(os.path.splitext(__file__)[0])
-VERSION = '8.6.9'
+VERSION = '9.0.4'
 
 # AsciiDoc global configuration file directory.
 # NOTE: CONF_DIR is "fixed up" by Makefile -- don't rename or change syntax.
@@ -44,16 +63,18 @@ ASCIIDOC = 'asciidoc'
 XSLTPROC = 'xsltproc'
 DBLATEX = 'dblatex'         # pdf generation.
 FOP = 'fop'                 # pdf generation (--fop option).
-W3M = 'w3m'                 # text generation.
-LYNX = 'lynx'               # text generation (if no w3m).
+W3M = 'w3m'                 # primary text file generator.
+LYNX = 'lynx'               # alternate text file generator.
 XMLLINT = 'xmllint'         # Set to '' to disable.
 EPUBCHECK = 'epubcheck'     # Set to '' to disable.
 # External executable default options.
 ASCIIDOC_OPTS = ''
+BACKEND_OPTS = ''
 DBLATEX_OPTS = ''
 FOP_OPTS = ''
+LYNX_OPTS = '-dump'
+W3M_OPTS = '-dump -cols 70 -T text/html -no-graph'
 XSLTPROC_OPTS = ''
-BACKEND_OPTS = ''
 
 ######################################################################
 # End of configuration file parameters.
@@ -66,59 +87,76 @@ BACKEND_OPTS = ''
 
 OPTIONS = None  # These functions read verbose and dry_run command options.
 
+
 def errmsg(msg):
-    sys.stderr.write('%s: %s\n' % (PROG,msg))
+    print('%s: %s\n' % (PROG, msg), file=sys.stderr)
+
 
 def warning(msg):
     errmsg('WARNING: %s' % msg)
 
+
 def infomsg(msg):
-    print '%s: %s' % (PROG,msg)
+    print('%s: %s' % (PROG, msg))
+
 
 def die(msg, exit_code=1):
     errmsg('ERROR: %s' % msg)
     sys.exit(exit_code)
 
+
 def trace():
     """Print traceback to stderr."""
     errmsg('-'*60)
     traceback.print_exc(file=sys.stderr)
     errmsg('-'*60)
 
+
 def verbose(msg):
     if OPTIONS.verbose or OPTIONS.dry_run:
         infomsg(msg)
 
+
 class AttrDict(dict):
     """
     Like a dictionary except values can be accessed as attributes i.e. obj.foo
     can be used in addition to obj['foo'].
-    If self._default has been set then it will be returned if a non-existant
+    If self._default has been set then it will be returned if a non-existent
     attribute is accessed (instead of raising an AttributeError).
     """
     def __getattr__(self, key):
         try:
             return self[key]
-        except KeyError, k:
-            if self.has_key('_default'):
+        except KeyError as k:
+            if '_default' in self:
                 return self['_default']
             else:
-                raise AttributeError, k
+                raise AttributeError from k
+
     def __setattr__(self, key, value):
         self[key] = value
+
     def __delattr__(self, key):
-        try: del self[key]
-        except KeyError, k: raise AttributeError, k
+        try:
+            del self[key]
+        except KeyError as k:
+            raise AttributeError from k
+
     def __repr__(self):
         return '<AttrDict ' + dict.__repr__(self) + '>'
+
     def __getstate__(self):
         return dict(self)
-    def __setstate__(self,value):
-        for k,v in value.items(): self[k]=v
+
+    def __setstate__(self, value):
+        for k, v in value.items():
+            self[k] = v
+
 
 def isexecutable(file_name):
     return os.path.isfile(file_name) and os.access(file_name, os.X_OK)
 
+
 def find_executable(file_name):
     '''
     Search for executable file_name in the system PATH.
@@ -137,32 +175,31 @@ def find_executable(file_name):
                 return os.path.realpath(f)
         return None
     if os.name == 'nt' and os.path.splitext(file_name)[1] == '':
-        for ext in ('.cmd','.bat','.exe'):
+        for ext in ('.cmd', '.bat', '.exe'):
             result = _find_executable(file_name + ext)
-            if result: break
+            if result:
+                break
     else:
         result = _find_executable(file_name)
     return result
 
-def write_file(filename, data, mode='w'):
-    f = open(filename, mode)
-    try:
-        f.write(data)
-    finally:
-        f.close()
 
-def read_file(filename, mode='r'):
-    f = open(filename, mode)
-    try:
+def write_file(filename, data, mode='w', encoding='utf-8'):
+    with open(filename, mode=mode, encoding=encoding) as f:
+        f.write(data)
+
+
+def read_file(filename, mode='r', encoding='utf-8'):
+    with open(filename, mode=mode, encoding=encoding) as f:
         return f.read()
-    finally:
-        f.close()
+
 
 def shell_cd(path):
     verbose('chdir %s' % path)
     if not OPTIONS.dry_run:
         os.chdir(path)
 
+
 def shell_makedirs(path):
     if os.path.isdir(path):
         return
@@ -170,11 +207,13 @@ def shell_makedirs(path):
     if not OPTIONS.dry_run:
         os.makedirs(path)
 
+
 def shell_copy(src, dst):
-    verbose('copying "%s" to "%s"' % (src,dst))
+    verbose('copying "%s" to "%s"' % (src, dst))
     if not OPTIONS.dry_run:
         shutil.copy(src, dst)
 
+
 def shell_rm(path):
     if not os.path.exists(path):
         return
@@ -182,6 +221,7 @@ def shell_rm(path):
     if not OPTIONS.dry_run:
         os.unlink(path)
 
+
 def shell_rmtree(path):
     if not os.path.isdir(path):
         return
@@ -189,6 +229,7 @@ def shell_rmtree(path):
     if not OPTIONS.dry_run:
         shutil.rmtree(path)
 
+
 def shell(cmd, raise_error=True):
     '''
     Execute command cmd in shell and return tuple
@@ -216,17 +257,18 @@ def shell(cmd, raise_error=True):
     stdout = stderr = subprocess.PIPE
     try:
         popen = subprocess.Popen(cmd, stdout=stdout, stderr=stderr,
-                shell=True, env=ENV)
-    except OSError, e:
+                                 shell=True, env=ENV, universal_newlines=True)
+    except OSError as e:
         die('failed: %s: %s' % (cmd, e))
     stdoutdata, stderrdata = popen.communicate()
     if OPTIONS.verbose:
-        print stdoutdata
-        print stderrdata
+        print(stdoutdata)
+        print(stderrdata)
     if popen.returncode != 0 and raise_error:
         die('%s returned non-zero exit status %d' % (cmd, popen.returncode))
     return (stdoutdata, stderrdata, popen.returncode)
 
+
 def find_resources(files, tagname, attrname, filter=None):
     '''
     Search all files and return a list of local URIs from attrname attribute
@@ -234,20 +276,22 @@ def find_resources(files, tagname, attrname, filter=None):
     Handles HTML open and XHTML closed tags.
     Non-local URIs are skipped.
     files can be a file name or a list of file names.
-    The filter function takes a dictionary of tag attributes and returns True if
-    the URI is to be included.
+    The filter function takes a dictionary of tag attributes and returns True
+    if the URI is to be included.
     '''
-    class FindResources(HTMLParser.HTMLParser):
+    class FindResources(HTMLParser):
         # Nested parser class shares locals with enclosing function.
         def handle_startendtag(self, tag, attrs):
             self.handle_starttag(tag, attrs)
+
         def handle_starttag(self, tag, attrs):
             attrs = dict(attrs)
             if tag == tagname and (filter is None or filter(attrs)):
                 # Accept only local URIs.
-                uri = urlparse.urlparse(attrs[attrname])
-                if uri[0] in ('','file') and not uri[1] and uri[2]:
+                uri = urlparse(attrs[attrname])
+                if uri[0] in ('', 'file') and not uri[1] and uri[2]:
                     result.append(uri[2])
+
     if isinstance(files, str):
         files = [files]
     result = []
@@ -256,20 +300,19 @@ def find_resources(files, tagname, attrname, filter=None):
         if OPTIONS.dry_run:
             continue
         parser = FindResources()
-        # HTMLParser has problems with non-ASCII strings.
-        # See http://bugs.python.org/issue3932
-        contents = read_file(filename)
-        mo = re.search(r'\A<\?xml.* encoding="(.*?)"', contents)
-        if mo:
-            encoding = mo.group(1)
-            parser.feed(contents.decode(encoding))
-        else:
-            parser.feed(contents)
+        with open(filename, 'rb') as open_file:
+            contents = open_file.read()
+        mo = re.search(b'\A<\?xml.* encoding="(.*?)"', contents)
+        if mo is None:
+            mo = re.search(br'<meta http\-equiv="Content\-Type" content="text\/html; charset=(.*?)">', contents)
+        contents = contents.decode(mo.group(1).decode('utf-8') if mo else 'utf-8')
+        parser.feed(contents)
         parser.close()
     result = list(set(result))   # Drop duplicate values.
     result.sort()
     return result
 
+
 # NOT USED.
 def copy_files(files, src_dir, dst_dir):
     '''
@@ -289,18 +332,20 @@ def copy_files(files, src_dir, dst_dir):
             shell_makedirs(dstdir)
             shell_copy(src, dst)
 
+
 def find_files(path, pattern):
     '''
     Return list of file names matching pattern in directory path.
     '''
     result = []
-    for (p,dirs,files) in os.walk(path):
+    for (p, dirs, files) in os.walk(path):
         for f in files:
             if fnmatch.fnmatch(f, pattern):
-                result.append(os.path.normpath(os.path.join(p,f)))
+                result.append(os.path.normpath(os.path.join(p, f)))
     return result
 
-def exec_xsltproc(xsl_file, xml_file, dst_dir, opts = ''):
+
+def exec_xsltproc(xsl_file, xml_file, dst_dir, opts=''):
     cwd = os.getcwd()
     shell_cd(dst_dir)
     try:
@@ -308,6 +353,7 @@ def exec_xsltproc(xsl_file, xml_file, dst_dir, opts = ''):
     finally:
         shell_cd(cwd)
 
+
 def get_source_options(asciidoc_file):
     '''
     Look for a2x command options in AsciiDoc source file.
@@ -339,14 +385,19 @@ def get_source_options(asciidoc_file):
     result = []
     if os.path.isfile(asciidoc_file):
         options = ''
-        f = open(asciidoc_file)
-        try:
+        with open(asciidoc_file, 'rb') as f:
+            line_number = 0
             for line in f:
-                mo = re.search(r'^//\s*a2x:', line)
+                line_number += 1
+                mo = re.search(b'^//\s*a2x:', line)
                 if mo:
-                    options += ' ' + line[mo.end():].strip()
-        finally:
-            f.close()
+                    try:
+                        options += ' ' + line[mo.end():].strip().decode('ascii')
+                    except UnicodeDecodeError as e:
+                        warning(
+                            "Could not decode option to %s " % e.encoding +
+                            "on line %s in %s" % (line_number, asciidoc_file)
+                        )
         parse_options()
     return result
 
@@ -367,10 +418,10 @@ class A2X(AttrDict):
         self.process_options()
         # Append configuration file options.
         self.asciidoc_opts += ' ' + ASCIIDOC_OPTS
-        self.dblatex_opts  += ' ' + DBLATEX_OPTS
-        self.fop_opts      += ' ' + FOP_OPTS
+        self.dblatex_opts += ' ' + DBLATEX_OPTS
+        self.fop_opts += ' ' + FOP_OPTS
         self.xsltproc_opts += ' ' + XSLTPROC_OPTS
-        self.backend_opts  += ' ' + BACKEND_OPTS
+        self.backend_opts += ' ' + BACKEND_OPTS
         # Execute to_* functions.
         if self.backend:
             self.to_backend()
@@ -435,14 +486,14 @@ class A2X(AttrDict):
         for f in conf_files:
             if os.path.isfile(f):
                 verbose('loading configuration file: %s' % f)
-                execfile(f, globals())
+                exec(open(f).read(), globals())
 
     def process_options(self):
         '''
         Validate and command options and set defaults.
         '''
         if not os.path.isfile(self.asciidoc_file):
-            die('missing SOURCE_FILE: %s' % self.asciidoc_file)
+            die('missing ASCIIDOC_FILE: %s' % self.asciidoc_file)
         self.asciidoc_file = os.path.abspath(self.asciidoc_file)
         if os.path.splitext(self.asciidoc_file)[1].lower() == '.xml':
             self.skip_asciidoc = True
@@ -454,24 +505,21 @@ class A2X(AttrDict):
             if not os.path.isdir(self.destination_dir):
                 die('missing --destination-dir: %s' % self.destination_dir)
             self.destination_dir = os.path.abspath(self.destination_dir)
-            if not self.format in ('chunked','epub','htmlhelp','xhtml'):
-                warning('--destination-dir option is only applicable to HTML based outputs')
+            if self.format not in ('chunked', 'epub', 'htmlhelp', 'xhtml', 'manpage'):
+                warning('--destination-dir option is only applicable to HTML and manpage based outputs')
         self.resource_dirs = []
         self.resource_files = []
         if self.resource_manifest:
             if not os.path.isfile(self.resource_manifest):
                 die('missing --resource-manifest: %s' % self.resource_manifest)
-            f = open(self.resource_manifest)
-            try:
+            with open(self.resource_manifest) as f:
                 for r in f:
                     self.resources.append(r.strip())
-            finally:
-                f.close()
         for r in self.resources:
             r = os.path.expanduser(r)
             r = os.path.expandvars(r)
             if r.endswith('/') or r.endswith('\\'):
-                if  os.path.isdir(r):
+                if os.path.isdir(r):
                     self.resource_dirs.append(r)
                 else:
                     die('missing resource directory: %s' % r)
@@ -483,8 +531,8 @@ class A2X(AttrDict):
             else:
                 self.resource_files.append(r)
         for p in (os.path.dirname(self.asciidoc), CONF_DIR):
-            for d in ('images','stylesheets'):
-                d = os.path.join(p,d)
+            for d in ('images', 'stylesheets'):
+                d = os.path.join(p, d)
                 if os.path.isdir(d):
                     self.resource_dirs.append(d)
         verbose('resource files: %s' % self.resource_files)
@@ -578,16 +626,21 @@ class A2X(AttrDict):
 
     def copy_resources(self, html_files, src_dir, dst_dir, resources=[]):
         '''
-        Search html_files for images and CSS resource URIs (html_files can be a
-        list of file names or a single file name).
+        Search html_files for images and CSS resource URIs (html_files can
+        be a list of file names or a single file name).
         Copy them from the src_dir to the dst_dir.
         If not found in src_dir then recursively search all specified
         resource directories.
-        Optional additional resources files can be passed in the resources list.
+        Optional additional resources files can be passed in the resources
+        list.
         '''
         resources = resources[:]
-        resources += find_resources(html_files, 'link', 'href',
-                        lambda attrs: attrs.get('type') == 'text/css')
+        resources += find_resources(
+            html_files,
+            'link',
+            'href',
+            lambda attrs: attrs.get('type') == 'text/css'
+        )
         resources += find_resources(html_files, 'img', 'src')
         resources += self.resource_files
         resources = list(set(resources))    # Drop duplicates.
@@ -710,7 +763,7 @@ class A2X(AttrDict):
             dst_dir = self.dst_path('.chunked')
         elif self.format == 'htmlhelp':
             dst_dir = self.dst_path('.htmlhelp')
-        if not 'base.dir ' in opts:
+        if 'base.dir ' not in opts:
             opts += ' --stringparam base.dir "%s/"' % os.path.basename(dst_dir)
         # Create content.
         shell_rmtree(dst_dir)
@@ -727,9 +780,9 @@ class A2X(AttrDict):
         '''
         opf_dir = os.path.dirname(opf_file)
         resource_files = []
-        for (p,dirs,files) in os.walk(os.path.dirname(opf_file)):
+        for (p, dirs, files) in os.walk(os.path.dirname(opf_file)):
             for f in files:
-                f = os.path.join(p,f)
+                f = os.path.join(p, f)
                 if os.path.isfile(f):
                     assert f.startswith(opf_dir)
                     f = '.' + f[len(opf_dir):]
@@ -785,21 +838,21 @@ class A2X(AttrDict):
         shell_cd(build_dir)
         try:
             if not self.dry_run:
-                zip = zipfile.ZipFile(epub_file, 'w')
+                zip_archive = zipfile.ZipFile(epub_file, 'w')
                 try:
                     # Create and add uncompressed mimetype file.
                     verbose('archiving: mimetype')
                     write_file('mimetype', 'application/epub+zip')
-                    zip.write('mimetype', compress_type=zipfile.ZIP_STORED)
+                    zip_archive.write('mimetype', compress_type=zipfile.ZIP_STORED)
                     # Compress all remaining files.
-                    for (p,dirs,files) in os.walk('.'):
+                    for (p, dirs, files) in os.walk('.'):
                         for f in files:
                             f = os.path.normpath(os.path.join(p,f))
                             if f != 'mimetype':
                                 verbose('archiving: %s' % f)
-                                zip.write(f, compress_type=zipfile.ZIP_DEFLATED)
+                                zip_archive.write(f, compress_type=zipfile.ZIP_DEFLATED)
                 finally:
-                    zip.close()
+                    zip_archive.close()
             verbose('created archive: %s' % epub_file)
         finally:
             shell_cd(cwd)
@@ -818,8 +871,8 @@ class A2X(AttrDict):
             shell('"%s" %s --conf-file "%s" -b html4 -a "a2x-format=%s" -o "%s" "%s"' %
                  (self.asciidoc, self.asciidoc_opts, self.asciidoc_conf_file('text.conf'),
                   self.format, html_file, self.asciidoc_file))
-            shell('"%s" -dump "%s" > "%s"' %
-                 (LYNX, html_file, text_file))
+            cmd = '"%s" %s "%s" > "%s"' % (LYNX, LYNX_OPTS, html_file, text_file)
+            shell(cmd)
         else:
             # Use w3m(1).
             self.to_docbook()
@@ -827,8 +880,8 @@ class A2X(AttrDict):
             opts = '%s --output "%s"' % (self.xsltproc_opts, html_file)
             exec_xsltproc(self.xsl_stylesheet(), docbook_file,
                     self.destination_dir, opts)
-            shell('"%s" -cols 70 -dump -T text/html -no-graph "%s" > "%s"' %
-                 (W3M, html_file, text_file))
+            cmd = '"%s" %s "%s" > "%s"' % (W3M, W3M_OPTS, html_file, text_file)
+            shell(cmd)
         if not self.keep_artifacts:
             shell_rm(html_file)
 
@@ -959,4 +1012,4 @@ if __name__ == '__main__':
         a2x.load_conf()
         a2x.execute()
     except KeyboardInterrupt:
-        exit(1)
+        sys.exit(1)
