@@ -1,17 +1,6 @@
-#!/usr/local/bin/python2.7
+#!/usr/bin/env python3
 
-import os, sys, subprocess
-from optparse import *
-
-__AUTHOR__ = "Gouichi Iisaka <iisaka51@gmail.com>"
-__VERSION__ = '1.1.4'
-
-class EApp(Exception):
-    '''Application specific exception.'''
-    pass
-
-class Application():
-    '''
+'''
 NAME
     graphviz2png - Converts textual graphviz notation to PNG file
 
@@ -61,50 +50,64 @@ LICENSE
     Copyright (C) 2008-2009 Gouichi Iisaka.
     Free use of this software is granted under the terms of
     the GNU General Public License (GPL).
-    '''
+'''
 
+import os
+import sys
+import subprocess
+import argparse
+
+__AUTHOR__ = "Gouichi Iisaka <iisaka51@gmail.com>"
+__VERSION__ = '1.1.5'
+
+
+class EApp(Exception):
+    '''Application specific exception.'''
+    pass
+
+
+class Application():
     def __init__(self, argv=None):
         # Run dot, get the list of supported formats. It's prefixed by some junk.
-        format_output = subprocess.Popen(["dot", "-T?"], stderr=subprocess.PIPE, stdout=subprocess.PIPE).communicate()[1]
-        # The junk contains : and ends with :. So we split it, then strip the final endline, then split the list for future usage.
+        format_output = subprocess.Popen(
+            ["dot", "-T?"],
+            stderr=subprocess.PIPE,
+            stdout=subprocess.PIPE
+        ).communicate()[1].decode('utf-8')
+        # The junk contains : and ends with :. So we split it, then strip the
+        # final endline, then split the list for future usage.
         supported_formats = format_output.split(": ")[2][:-1].split(" ")
 
         if not argv:
             argv = sys.argv
 
-        self.usage = '%prog [options] inputfile'
+        self.usage = '%(prog)s [options] infile'
         self.version = 'Version: %s\n' % __VERSION__
         self.version += 'Copyright(c) 2008-2009: %s\n' % __AUTHOR__
 
-        self.option_list = [
-            Option("-o", "--outfile", action="store",
-                dest="outfile",
-                help="Output file"),
-            Option("-L", "--layout", action="store",
-                dest="layout", default="dot", type="choice",
-                choices=['dot','neato','twopi','circo','fdp'],
-                help="Layout type. LAYOUT=<dot|neato|twopi|circo|fdp>"),
-            Option("-F", "--format", action="store",
-                dest="format", default="png", type="choice",
-                choices=supported_formats,
-                help="Format type. FORMAT=<" + "|".join(supported_formats) + ">"),
-            Option("--debug", action="store_true",
-                dest="do_debug",
-                help=SUPPRESS_HELP),
-            Option("-v", "--verbose", action="store_true",
-                dest="do_verbose", default=False,
-                help="verbose output"),
-            ]
-
-        self.parser = OptionParser( usage=self.usage, version=self.version,
-                                    option_list=self.option_list)
-        (self.options, self.args) = self.parser.parse_args()
-
-        if len(self.args) != 1:
-            self.parser.print_help()
-            sys.exit(1)
-
-        self.options.infile = self.args[0]
+        self.parser = argparse.ArgumentParser(usage=self.usage)
+        self.parser.add_argument(
+            "-o", "--outfile", action="store", dest="outfile", help="Output file"
+        )
+        self.parser.add_argument(
+            "-L", "--layout", action="store", dest="layout", default="dot",
+            choices=['dot', 'neato', 'twopi', 'circo', 'fdp'], help="Layout type"
+        )
+        self.parser.add_argument(
+            "-F", "--format", action="store", dest="format", default="png",
+            choices=supported_formats, help="Format type"
+        )
+        self.parser.add_argument(
+            "--debug", action="store_true", dest="do_debug", help=argparse.SUPPRESS
+        )
+        self.parser.add_argument(
+            "-v", "--verbose",
+            action="store_true", dest="do_verbose", default=False,
+            help="verbose output"
+        )
+        self.parser.add_argument("infile", action="store", help="Input file")
+        self.parser.add_argument('--version', action='version', version=self.version)
+        self.options = self.parser.parse_args()
 
     def systemcmd(self, cmd):
         if self.options.do_verbose:
@@ -113,7 +116,7 @@ LICENSE
         else:
             cmd += ' 2>%s' % os.devnull
         if os.system(cmd):
-            raise EApp, 'failed command: %s' % cmd
+            raise EApp('failed command: %s' % cmd)
 
     def graphviz2png(self, infile, outfile):
         '''Convert Graphviz notation in file infile to
@@ -123,14 +126,17 @@ LICENSE
         outdir = os.path.dirname(outfile)
 
         if not os.path.isdir(outdir):
-            raise EApp, 'directory does not exist: %s' % outdir
+            raise EApp('directory does not exist: %s' % outdir)
 
-        basefile = os.path.splitext(outfile)[0]
         saved_cwd = os.getcwd()
         os.chdir(outdir)
         try:
             cmd = '%s -T%s "%s" > "%s"' % (
-                  self.options.layout, self.options.format, infile, outfile)
+                self.options.layout,
+                self.options.format,
+                infile,
+                outfile
+            )
             self.systemcmd(cmd)
         finally:
             os.chdir(saved_cwd)
@@ -142,6 +148,7 @@ LICENSE
         if self.options.format == '':
             self.options.format = 'png'
 
+        infile = self.options.infile
         if self.options.infile == '-':
             if self.options.outfile is None:
                 sys.stderr.write('OUTFILE must be specified')
@@ -151,7 +158,7 @@ LICENSE
             open(infile, 'w').writelines(lines)
 
         if not os.path.isfile(infile):
-            raise EApp, 'input file does not exist: %s' % infile
+            raise EApp('input file does not exist: %s' % infile)
 
         if self.options.outfile is None:
             outfile = os.path.splitext(infile)[0] + '.png'
@@ -164,6 +171,7 @@ LICENSE
         if self.options.infile == '-':
             sys.stdout.write(' ')
 
+
 if __name__ == "__main__":
     app = Application()
     app.run()
